
FRONTEND_DIR = 'frontend'
FRONTEND_URL_PREFIX = 'frontend'

APP_NAME = 'qtoggleserver'
APP_PRETTY_NAME = 'qToggle Remote'
APP_DESCRIPTION = 'Control your qToggle devices'

BROWSER_BACKGROUND_COLOR = '#444444'
BROWSER_THEME_COLOR = '#62abea'
